use std::collections::{HashSet, VecDeque};
use std::sync::Arc;

use idb::{
    Database, DatabaseEvent, Factory, KeyPath, ObjectStore, ObjectStoreParams, Query, TransactionMode,
};
use offline_web_model::Reference;
use serde::{Deserialize, Serialize};
use wasm_bindgen::JsValue;

use crate::ReferenceStore;
use crate::StoreError;

const DB_NAME: &str = "offline_web_store";
const DB_VERSION: u32 = 1;
const REF_ENTRIES_STORE: &str = "ref_entries";
const REF_DEPENDENCIES_STORE: &str = "ref_dependencies";
const CONTENT_STORE: &str = "content_store";

#[derive(Serialize, Deserialize, Clone)]
struct RefEntry {
    id: String,
    content_address: Option<String>,
    name: String,
}

#[derive(Serialize, Deserialize, Clone)]
struct RefDependency {
    parent_id: String,
    dependent_id: String,
}

#[derive(Serialize, Deserialize, Clone)]
struct ContentEntry {
    content_address: String,
    content: Vec<u8>,
}

pub struct IndexedDbReferenceStore {
    db: Database,
}

impl IndexedDbReferenceStore {
    pub async fn new() -> Result<Self, StoreError> {
        let factory = Factory::new()?;

        let mut open_request = factory.open(DB_NAME, Some(DB_VERSION))?;

        // Set up database upgrade handler
        open_request.on_upgrade_needed(|event| {
            let db = event.database().expect("Failed to get indexeddb database");

            // Create ref_entries object store
            if !db
                .store_names()
                .iter()
                .any(|name| name == REF_ENTRIES_STORE)
            {
                let ref_store = db.create_object_store(REF_ENTRIES_STORE, ObjectStoreParams::new())
                    .expect(&format!(
                        "Failed to create ref_entries store: {:?}",
                        REF_ENTRIES_STORE
                    ));
                
                // Create name index for get_graph method
                ref_store.create_index("name", KeyPath::new_single("name"), None)
                    .expect("Failed to create name index");
            }

            // Create ref_dependencies object store
            if !db
                .store_names()
                .iter()
                .any(|name| name == REF_DEPENDENCIES_STORE)
            {
                db.create_object_store(REF_DEPENDENCIES_STORE, ObjectStoreParams::new())
                    .expect(&format!(
                        "Failed to create ref_dependencies store: {:?}",
                        REF_DEPENDENCIES_STORE
                    ));
            }

            // Create content_store object store
            if !db.store_names().iter().any(|name| name == CONTENT_STORE) {
                db.create_object_store(CONTENT_STORE, ObjectStoreParams::new())
                    .expect(&format!(
                        "Failed to create content_store: {:?}",
                        CONTENT_STORE
                    ));
            }
        });

        let db = open_request.await?;

        Ok(Self { db })
    }


    async fn clear_dependencies_sync(
        &self,
        dep_store: &ObjectStore,
        parent_id: &str,
    ) -> Result<(), StoreError> {
        let cursor_request = dep_store.open_cursor(None, None)?;
        let mut keys_to_delete: Vec<String> = Vec::new();
        
        if let Ok(Some(cursor)) = cursor_request.await {
            loop {
                let should_continue = match cursor.value() {
                    Ok(value) => {
                        if let Some(dep_str) = value.as_string() {
                            if let Ok(dep_entry) = serde_json::from_str::<RefDependency>(&dep_str) {
                                if dep_entry.parent_id == parent_id {
                                    if let Ok(key) = cursor.key() {
                                        if let Some(key_str) = key.as_string() {
                                            keys_to_delete.push(key_str);
                                        }
                                    }
                                }
                            }
                        }
                        true
                    }
                    Err(_) => false,
                };
                
                if !should_continue {
                    break;
                }
                
                // Try to continue to next item
                match cursor.advance(1) {
                    Ok(advance_request) => {
                        if advance_request.await.is_err() {
                            break;
                        }
                        if cursor.value().is_err() {
                            break;
                        }
                    }
                    Err(_) => break,
                }
            }
        }

        // Delete the found keys
        for key in keys_to_delete {
            let delete_request = dep_store.delete(JsValue::from_str(&key))?;
            delete_request.await?;
        }

        Ok(())
    }

    async fn get_dependents(&self, parent_id: &str) -> Result<Vec<Arc<Reference>>, StoreError> {
        // Use a completely iterative approach to build the dependency tree
        let mut all_refs = std::collections::HashMap::new();
        let mut dependency_map = std::collections::HashMap::<String, Vec<String>>::new();
        let mut to_process = std::collections::VecDeque::new();
        let mut processed = std::collections::HashSet::new();
        
        // Start with direct children of the parent
        let direct_children = self.get_direct_children(parent_id).await?;
        for child_id in direct_children.clone() {
            to_process.push_back(child_id);
        }
        
        // Process all references iteratively to build both refs map and dependency map
        while let Some(ref_id) = to_process.pop_front() {
            if processed.contains(&ref_id) {
                continue;
            }
            processed.insert(ref_id.clone());
            
            // Get the reference without dependents first
            if let Ok(reference) = self.get_reference_without_dependents(&ref_id).await {
                all_refs.insert(ref_id.clone(), reference);
                
                // Get its direct children and add them to the processing queue
                if let Ok(children) = self.get_direct_children(&ref_id).await {
                    dependency_map.insert(ref_id.clone(), children.clone());
                    for child_id in children {
                        to_process.push_back(child_id);
                    }
                } else {
                    dependency_map.insert(ref_id.clone(), Vec::new());
                }
            }
        }
        
        // Now build the dependency tree iteratively using the maps
        let mut result = Vec::new();
        for child_id in direct_children {
            if let Some(reference) = self.build_full_reference(&child_id, &all_refs, &dependency_map) {
                result.push(Arc::new(reference));
            }
        }
        
        Ok(result)
    }
    
    fn build_full_reference(
        &self,
        target_id: &str,
        all_refs: &std::collections::HashMap<String, Reference>,
        dependency_map: &std::collections::HashMap<String, Vec<String>>
    ) -> Option<Reference> {
        let mut built_refs = std::collections::HashMap::<String, Reference>::new();
        let mut to_build = std::collections::VecDeque::new();
        let mut processed = std::collections::HashSet::new();
        
        // Start from leaves and work up
        to_build.push_back(target_id.to_string());
        
        while let Some(ref_id) = to_build.pop_back() {
            if processed.contains(&ref_id) {
                continue;
            }
            
            let base_ref = all_refs.get(&ref_id)?;
            let children = dependency_map.get(&ref_id).cloned().unwrap_or_default();
            
            // Check if all children are already built
            let all_children_built = children.iter().all(|child_id| built_refs.contains_key(child_id));
            
            if all_children_built {
                // Build this reference with its dependents
                let mut dependents = Vec::new();
                for child_id in &children {
                    if let Some(child_ref) = built_refs.get(child_id) {
                        dependents.push(Arc::new(child_ref.clone()));
                    }
                }
                
                let full_ref = Reference {
                    id: base_ref.id.clone(),
                    content_address: base_ref.content_address.clone(),
                    name: base_ref.name.clone(),
                    dependents,
                };
                
                built_refs.insert(ref_id.clone(), full_ref);
                processed.insert(ref_id);
            } else {
                // Add children to be built first, then re-add this ref
                to_build.push_front(ref_id);
                for child_id in children {
                    if !processed.contains(&child_id) && !built_refs.contains_key(&child_id) {
                        to_build.push_back(child_id);
                    }
                }
            }
        }
        
        built_refs.remove(target_id)
    }

    async fn find_reference_by_name(&self, name: &str) -> Result<Option<RefEntry>, StoreError> {
        let transaction = self
            .db
            .transaction(&[REF_ENTRIES_STORE], TransactionMode::ReadOnly)?;

        let ref_store = transaction.object_store(REF_ENTRIES_STORE)?;
        
        // Use cursor to search through all references
        let cursor_request = ref_store.open_cursor(None, None)?;
        
        if let Ok(Some(cursor)) = cursor_request.await {
            loop {
                let should_continue = match cursor.value() {
                    Ok(value) => {
                        if let Some(ref_str) = value.as_string() {
                            if let Ok(ref_entry) = serde_json::from_str::<RefEntry>(&ref_str) {
                                if ref_entry.name == name {
                                    return Ok(Some(ref_entry));
                                }
                            }
                        }
                        true
                    }
                    Err(_) => false,
                };
                
                if !should_continue {
                    break;
                }
                
                // Try to continue to next item
                match cursor.advance(1) {
                    Ok(advance_request) => {
                        if advance_request.await.is_err() {
                            break;
                        }
                        if cursor.value().is_err() {
                            break;
                        }
                    }
                    Err(_) => break,
                }
            }
        }
        
        Ok(None)
    }

    async fn get_reference_without_dependents(&self, id: &str) -> Result<Reference, StoreError> {
        let transaction = self
            .db
            .transaction(&[REF_ENTRIES_STORE], TransactionMode::ReadOnly)?;

        let ref_store = transaction.object_store(REF_ENTRIES_STORE)?;

        let request = ref_store.get(Query::Key(JsValue::from_str(id)))?;
        let value = request.await?;

        if let Some(value) = value {
            if let Some(ref_str) = value.as_string() {
                if let Ok(ref_entry) = serde_json::from_str::<RefEntry>(&ref_str) {
                    return Ok(Reference {
                        id: ref_entry.id,
                        content_address: ref_entry.content_address,
                        name: ref_entry.name,
                        dependents: Vec::new(), // No dependents in this method
                    });
                }
            }
        }

        Err(StoreError::NoSuchReference)
    }

    async fn get_direct_children(&self, parent_id: &str) -> Result<Vec<String>, StoreError> {
        let transaction = self
            .db
            .transaction(&[REF_DEPENDENCIES_STORE], TransactionMode::ReadOnly)?;

        let dep_store = transaction.object_store(REF_DEPENDENCIES_STORE)?;
        
        let mut children = Vec::new();
        let cursor_request = dep_store.open_cursor(None, None)?;
        
        if let Ok(Some(cursor)) = cursor_request.await {
            loop {
                let should_continue = match cursor.value() {
                    Ok(value) => {
                        if let Some(dep_str) = value.as_string() {
                            if let Ok(dep_entry) = serde_json::from_str::<RefDependency>(&dep_str) {
                                if dep_entry.parent_id == parent_id {
                                    children.push(dep_entry.dependent_id);
                                }
                            }
                        }
                        true
                    }
                    Err(_) => false,
                };
                
                if !should_continue {
                    break;
                }
                
                // Try to continue to next item
                match cursor.advance(1) {
                    Ok(advance_request) => {
                        // Wait for the advance operation, but if it fails, we're done
                        if advance_request.await.is_err() {
                            break;
                        }
                        // Check if cursor is still valid
                        if cursor.value().is_err() {
                            break;
                        }
                    }
                    Err(_) => break,
                }
            }
        }
        
        Ok(children)
    }
}

impl ReferenceStore for IndexedDbReferenceStore {
    async fn get_reference(&self, id: &str) -> Result<Reference, StoreError> {
        let mut reference = self.get_reference_without_dependents(id).await?;
        reference.dependents = self.get_dependents(id).await?;
        Ok(reference)
    }

    async fn get_content_for_reference(&self, reference: Reference) -> Result<String, StoreError> {
        if let Some(content_address) = &reference.content_address {
            let transaction = self
                .db
                .transaction(&[CONTENT_STORE], TransactionMode::ReadOnly)
                .map_err(|e| {
                    StoreError::StorageError(Box::new(std::io::Error::new(
                        std::io::ErrorKind::Other,
                        format!("Failed to create transaction: {:?}", e),
                    )))
                })?;

            let content_store = transaction.object_store(CONTENT_STORE).map_err(|e| {
                StoreError::StorageError(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    format!("Failed to get content_store: {:?}", e),
                )))
            })?;

            let request = content_store
                .get(Query::Key(JsValue::from_str(content_address)))
                .map_err(|e| {
                    StoreError::StorageError(Box::new(std::io::Error::new(
                        std::io::ErrorKind::Other,
                        format!("Failed to get content: {:?}", e),
                    )))
                })?;

            let value = request.await.map_err(|e| {
                StoreError::StorageError(Box::new(std::io::Error::new(
                    std::io::ErrorKind::Other,
                    format!("Failed to await content: {:?}", e),
                )))
            })?;

            if let Some(value) = value {
                if let Some(content_str) = value.as_string() {
                    if let Ok(content_entry) = serde_json::from_str::<ContentEntry>(&content_str) {
                        return String::from_utf8(content_entry.content)
                            .map_err(|e| StoreError::StorageError(Box::new(e)));
                    }
                }
            }

            Err(StoreError::NoSuchContentAddress)
        } else {
            Err(StoreError::NoSuchContentAddress)
        }
    }

    async fn get_graph(&self, root_name: &str) -> Result<Vec<Reference>, StoreError> {
        let mut visited = HashSet::new();
        let mut result = Vec::new();
        let mut queue = VecDeque::new();

        queue.push_back(root_name.to_string());

        while let Some(current_name) = queue.pop_front() {
            if visited.contains(&current_name) {
                continue;
            }

            visited.insert(current_name.clone());

            // Find reference by name - use a cursor to search through all references
            if let Ok(reference_opt) = self.find_reference_by_name(&current_name).await {
                if let Some(ref_entry) = reference_opt {
                    let dependents = self.get_dependents(&ref_entry.id).await?;

                    let reference = Reference {
                        id: ref_entry.id,
                        content_address: ref_entry.content_address,
                        name: ref_entry.name,
                        dependents: dependents.clone(),
                    };

                    result.push(reference);

                    // Add dependent names to queue
                    for dependent in dependents {
                        if !visited.contains(&dependent.name) {
                            queue.push_back(dependent.name.clone());
                        }
                    }
                }
            }
        }

        Ok(result)
    }
    
    async fn store_reference(&self, reference: &Reference) -> Result<(), StoreError> {
        let transaction = self
            .db
            .transaction(
                &[REF_ENTRIES_STORE, REF_DEPENDENCIES_STORE],
                TransactionMode::ReadWrite,
            )?;

        let ref_store = transaction.object_store(REF_ENTRIES_STORE)?;
        let dep_store = transaction.object_store(REF_DEPENDENCIES_STORE)?;

        // Store the reference entry
        let ref_entry = RefEntry {
            id: reference.id.clone(),
            content_address: reference.content_address.clone(),
            name: reference.name.clone(),
        };

        let ref_value = serde_json::to_string(&ref_entry)?;
        let ref_request = ref_store.put(
            &JsValue::from_str(&ref_value),
            Some(&JsValue::from_str(&reference.id)),
        )?;
        ref_request.await?;

        // Clear existing dependencies for this reference
        self.clear_dependencies_sync(&dep_store, &reference.id).await?;

        // Store new dependencies
        for dependent in &reference.dependents {
            let dep_entry = RefDependency {
                parent_id: reference.id.clone(),
                dependent_id: dependent.id.clone(),
            };

            let dep_value = serde_json::to_string(&dep_entry)?;
            let dep_key = format!("{}:{}", reference.id, dependent.id);
            
            let dep_request = dep_store.put(
                &JsValue::from_str(&dep_value),
                Some(&JsValue::from_str(&dep_key)),
            )?;
            dep_request.await?;
        }

        let commit_request = transaction.commit()?;
        commit_request.await?;

        Ok(())
    }
    
    async fn store_content(&self, content_address: &str, content: &[u8]) -> Result<(), StoreError> {
        let transaction = self
            .db
            .transaction(&[CONTENT_STORE], TransactionMode::ReadWrite)?;

        let content_store = transaction.object_store(CONTENT_STORE)?;

        let content_entry = ContentEntry {
            content_address: content_address.to_string(),
            content: content.to_vec(),
        };

        let content_value = serde_json::to_string(&content_entry)?;
        let put_request = content_store.put(
            &JsValue::from_str(&content_value),
            Some(&JsValue::from_str(content_address)),
        )?;
        put_request.await?;

        let commit_request = transaction.commit()?;
        commit_request.await?;

        Ok(())
    }
}

